@@ -1,5 +1,6 @@
-from struct import Struct
-import sqlite3
+from itertools import groupby
+from struct import Struct, unpack
+from sqlite3 import connect
 
 
 def stream_sqlite(sqlite_chunks, chunk_size=65536):
@@ -11,34 +12,13 @@ def stream_sqlite(sqlite_chunks, chunk_size=65536):
     signed_char = Struct('b')
     unsigned_short = Struct('>H')
     unsigned_long = Struct('>L')
+    table_header = Struct('>HHHB')
 
-    def get_byte_readers(iterable):
-        # Return functions to return bytes from the iterable
-        # - _yield_all: yields chunks as they come up (often for a "body")
-        # - _yield_num: yields chunks as the come up, up to a fixed number of bytes
-        # - _get_num: returns a single `bytes` of a given length
-
+    def get_byte_reader(iterable):
         chunk = b''
         offset = 0
         it = iter(iterable)
 
-        def _yield_all():
-            nonlocal chunk
-            nonlocal offset
-
-            while True:
-                if not chunk:
-                    try:
-                        chunk = next(it)
-                    except StopIteration:
-                        break
-                prev_offset = offset
-                prev_chunk = chunk
-                to_yield = min(len(chunk) - offset, chunk_size)
-                offset = (offset + to_yield) % len(chunk)
-                chunk = chunk if offset else b''
-                yield prev_chunk[prev_offset:prev_offset + to_yield]
-
         def _yield_num(num):
             nonlocal chunk
             nonlocal offset
@@ -48,7 +28,7 @@ def stream_sqlite(sqlite_chunks, chunk_size=65536):
                     try:
                         chunk = next(it)
                     except StopIteration:
-                        raise ValueError('Fewer bytes than expected in SQLite file') from None
+                        raise ValueError('Fewer bytes than expected in SQLite stream') from None
                 prev_offset = offset
                 prev_chunk = chunk
                 to_yield = min(num, len(chunk) - offset, chunk_size)
@@ -60,7 +40,7 @@ def stream_sqlite(sqlite_chunks, chunk_size=65536):
         def _get_num(num):
             return b''.join(chunk for chunk in _yield_num(num))
 
-        return _yield_all, _yield_num, _get_num
+        return _get_num
 
     def get_chunk_readers(chunk, p=0):
         # Set of functions to read a chunk of bytes, which it itself made of
@@ -94,51 +74,48 @@ def stream_sqlite(sqlite_chunks, chunk_size=65536):
 
         return _get_num, _get_varint
 
-    def type_length(serial_type):
-        return \
-            1 if serial_type == 1 else \
-            int((serial_type - 12)/2) if serial_type >= 12 and serial_type % 2 == 0 else \
-            int((serial_type - 13)/2) if serial_type >= 13 and serial_type % 2 == 1 else \
-            None
+    def yield_page_nums_pages_readers(get_bytes):
+        header = get_bytes(100)
 
-    def parse_serial_value(serial_type, raw):
-        if serial_type == 1:
-            return signed_char.unpack(raw)[0]
-        else:
-            return raw
+        if header[:16] != b'SQLite format 3\0':
+            raise ValueError('SQLite header not found at start of stream')
 
-    yield_all, yield_num, get_num = get_byte_readers(sqlite_chunks)
+        page_size, = unsigned_short.unpack(header[16:18])
+        page_size = 65536 if page_size == 1 else page_size
+        num_pages_expected, = unsigned_long.unpack(header[28:32])
 
-    header = get_num(100)
-
-    if header[:16] != b'SQLite format 3\0':
-        raise ValueError('SQLite header not found at start of stream')
-
-    page_size, = unsigned_short.unpack(header[16:18])
-    page_size = 65536 if page_size == 1 else page_size
-    num_pages_expected, = unsigned_long.unpack(header[28:32])
-
-    def yield_page_nums_pages_readers(page_size, num_pages_expected):
-        page_bytes = header + get_num(page_size - 100)
+        page_bytes = header + get_bytes(page_size - 100)
         page_reader, _ = get_chunk_readers(page_bytes)
         page_reader(100)
 
         yield 1, page_bytes, page_reader
 
         for page_num in range(2, num_pages_expected + 1):
-            page_bytes = get_num(page_size)
+            page_bytes = get_bytes(page_size)
             page_reader, _ = get_chunk_readers(page_bytes)
 
             yield page_num, page_bytes, page_reader
 
-    def yield_tables(page_nums_pages_readers):
-        cached_pages = {}
-        known_table_pages = {
+    def yield_table_pages(page_nums_pages_readers):
+        page_buffer = {}
+        page_tables = {
             1: 'sqlite_schema',
         }
         master_table = {}
 
-        def _yield_leaf_table_cells(page_bytes, pointers):
+        def type_length(serial_type):
+            return \
+                1 if serial_type == 1 else \
+                int((serial_type - 12)/2) if serial_type >= 12 and serial_type % 2 == 0 else \
+                int((serial_type - 13)/2) if serial_type >= 13 and serial_type % 2 == 1 else \
+                None
+
+        def parse_serial_value(serial_type, raw):
+            return \
+                signed_char.unpack(raw)[0] if serial_type == 1 else \
+                raw
+
+        def yield_leaf_table_cells(page_bytes, pointers):
             for pointer in pointers:
                 cell_num_reader, cell_varint_reader = get_chunk_readers(page_bytes, pointer)
 
@@ -159,13 +136,13 @@ def stream_sqlite(sqlite_chunks, chunk_size=65536):
                     for serial_type in serial_types
                 ]
 
-        def _yield_interior_table_cells(page_bytes, pointers):
+        def yield_interior_table_cells(page_bytes, pointers):
             for pointer in pointers:
                 cell_num_reader, cell_varint_reader = get_chunk_readers(page_bytes, pointer)
                 page_number, =  unsigned_long.unpack(cell_num_reader(4))
                 yield page_number
 
-        def _get_master_table(master_cells):
+        def get_master_table(master_cells):
 
             def schema(cur, table_name, sql):
                 cur.execute(sql)
@@ -174,7 +151,7 @@ def stream_sqlite(sqlite_chunks, chunk_size=65536):
                 cols = [d[0] for d in cur.description]
                 return [{col: row[i] for i, col in enumerate(cols)} for row in rows]
 
-            with sqlite3.connect(':memory:') as con:
+            with connect(':memory:') as con:
                 cur = con.cursor()
 
                 return [
@@ -187,62 +164,68 @@ def stream_sqlite(sqlite_chunks, chunk_size=65536):
                     if cell[0] == b'table'
                 ]
 
-        def cache_or_process(table_name, page_num):
+        def process_if_buffered_or_remember(table_name, page_num):
             try:
-                cached_page_bytes, cached_page_reader = cached_pages.pop(page_num)
+                page_bytes, page_reader = page_buffer.pop(page_num)
             except KeyError:
-                known_table_pages[page_num] = table_name
+                page_tables[page_num] = table_name
             else:
-                yield from process_table_page(table_name, cached_page_bytes, cached_page_reader)
+                yield from process_table_page(table_name, page_bytes, page_reader)
 
         def process_table_page(table_name, page_bytes, page_reader):
             page_type = page_reader(1)
             first_free_block, num_cells, cell_content_start, num_frag_free = \
-                Struct('>HHHB').unpack(page_reader(7))
+                table_header.unpack(page_reader(7))
             cell_content_start = 65536 if cell_content_start == 0 else cell_content_start
             right_most_pointer, = \
                 unsigned_long.unpack(page_reader(4)) if page_type == INTERIOR_TABLE else \
                 (None,)
 
-            pointers = Struct('>{}H'.format(num_cells)).unpack(page_reader(num_cells * 2))
+            pointers = unpack('>{}H'.format(num_cells), page_reader(num_cells * 2))
 
             if page_type == LEAF_TABLE and table_name == 'sqlite_schema':
-                for row in _get_master_table(_yield_leaf_table_cells(page_bytes, pointers)):
+                for row in get_master_table(yield_leaf_table_cells(page_bytes, pointers)):
                     master_table[row['name']] = row['info']
-                    yield from cache_or_process(row['name'], row['root_page'])
+                    yield from process_if_buffered_or_remember(row['name'], row['root_page'])
 
             elif page_type == LEAF_TABLE:
                 table_info = master_table[table_name]
-                yield table_name, table_info, [
+                yield table_name, table_info, (
                     {
                         table_info[i]['name']: value
                         for i, value in enumerate(cell)
                     }
-                    for cell in _yield_leaf_table_cells(page_bytes, pointers)
-                ]
+                    for cell in yield_leaf_table_cells(page_bytes, pointers)
+                )
 
             elif page_type == INTERIOR_TABLE:
-                for page_num in _yield_interior_table_cells(page_bytes, pointers):
-                    yield from cache_or_process(table_name, page_num)
-                yield from cache_or_process(table_name, right_most_pointer)
+                for page_num in yield_interior_table_cells(page_bytes, pointers):
+                    yield from process_if_buffered_or_remember(table_name, page_num)
+                yield from process_if_buffered_or_remember(table_name, right_most_pointer)
 
             else:
-                raise Exception('Unhandled page type')
+                raise ValueError('Unhandled page type in SQLite stream')
 
         for page_num, page_bytes, page_reader in page_nums_pages_readers:
             try:
-                table_name = known_table_pages.pop(page_num)
+                table_name = page_tables.pop(page_num)
             except KeyError:
-                cached_pages[page_num] = (page_bytes, page_reader)
+                page_buffer[page_num] = (page_bytes, page_reader)
             else:
                 yield from process_table_page(table_name, page_bytes, page_reader)
 
-    page_nums_pages_readers = yield_page_nums_pages_readers(page_size, num_pages_expected)
-    yield from yield_tables(page_nums_pages_readers)
+    def group_by_table(table_pages):
+        grouped_by_name = groupby(
+            table_pages, key=lambda name_info_pages: (name_info_pages[0], name_info_pages[1]))
 
-    extra = False
-    for _ in yield_all():
-        extra = True
+        def _rows(single_table_pages):
+            for _, _, table_rows in single_table_pages:
+                yield from table_rows
 
-    if extra:
-        raise ValueError('More bytes than expected in SQLite file')
+        for (name, info), single_table_pages in grouped_by_name:
+            yield name, info, _rows(single_table_pages)
+
+    get_bytes = get_byte_reader(sqlite_chunks)
+    page_nums_pages_readers = yield_page_nums_pages_readers(get_bytes)
+    table_pages = yield_table_pages(page_nums_pages_readers)
+    yield from group_by_table(table_pages)
